import traceback
from socketserver import BaseRequestHandler

from ldap import LDAPMessage, ShallowLDAPMessage, BindRequest, BindResponse, SearchRequest, SearchResultEntry, PartialAttribute, SearchResultDone, UnbindRequest, LDAPResultCode, IncompleteBERError, SimpleAuthentication, ModifyRequest, ModifyResponse, AddRequest, AddResponse, DelRequest, DelResponse, ModifyDNRequest, ModifyDNResponse, CompareRequest, CompareResponse, AbandonRequest, ExtendedRequest, ExtendedResponse, PasswdModifyRequestValue, PasswdModifyResponseValue

class LDAPError(Exception):
	def __init__(self, code=LDAPResultCode.other, message=''):
		self.code = code
		self.message = message

class LDAPOperationsError(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.operationsError, message)

class LDAPProtocolError(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.protocolError, message)

class LDAPInvalidCredentials(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.invalidCredentials, message)

class LDAPInsufficientAccessRights(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.insufficientAccessRights, message)

class LDAPUnwillingToPerform(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.unwillingToPerform, message)

class LDAPAuthMethodNotSupported(LDAPError):
	def __init__(self, message=''):
		super().__init__(LDAPResultCode.authMethodNotSupported, message)

class LDAPRequestHandler(BaseRequestHandler):
	def handle_bind(self, req):
		op = req.protocolOp
		if op.version != 3:
			raise LDAPProtocolError('Unsupported protocol version')
		if isinstance(op.authentication, SimpleAuthentication):
			self.bind_object = self.do_bind(op.name, op.authentication.password)
		else:
			raise LDAPAuthMethodNotSupported()
		self.send_msg(LDAPMessage(req.messageID, BindResponse(LDAPResultCode.success)))

	def do_bind(self, user, password):
		raise LDAPInvalidCredentials()

	def handle_search(self, req):
		search = req.protocolOp
		for dn, attributes in self.do_search(search.baseObject, search.scope, search.filter):
			attributes = [PartialAttribute(name, values) for name, values in attributes.items()]
			self.send_msg(LDAPMessage(req.messageID, SearchResultEntry(dn, attributes)))
		self.send_msg(LDAPMessage(req.messageID, SearchResultDone(LDAPResultCode.success)))

	def do_search(self, baseobj, scope, filter):
		yield from []

	def handle_unbind(self, req):
		self.keep_running = False

	def handle_extended(self, req):
		op = req.protocolOp
		if op.requestName == '1.3.6.1.4.1.1466.20037':
			# StartTLS (RFC 4511)
			raise LDAPProtocolError()
		elif op.requestName == '1.3.6.1.4.1.4203.1.11.1':
			# Password Modify Extended Operation (RFC 3062)
			newpw = None
			if op.requestValue is None:
				newpw = self.do_passwd()
			else:
				decoded, _ = PasswdModifyRequestValue.from_ber(op.requestValue)
				newpw = self.do_passwd(decoded.userIdentity, decoded.oldPasswd, decoded.newPasswd)
			if newpw is None:
				self.send_msg(LDAPMessage(req.messageID, ExtendedResponse(LDAPResultCode.success)))
			else:
				encoded = PasswdModifyResponseValue.to_ber(PasswdModifyResponseValue(newpw))
				self.send_msg(LDAPMessage(req.messageID, ExtendedResponse(LDAPResultCode.success, responseValue=encoded)))
		else:
			raise LDAPProtocolError()

	def do_passwd(self, user=None, oldpasswd=None, newpasswd=None):
		raise LDAPUnwillingToPerform('Password change is not supported')

	def handle_message(self, data):
		handlers = {
			BindRequest: (self.handle_bind, BindResponse),
			UnbindRequest: (self.handle_unbind, None),
			SearchRequest: (self.handle_search, SearchResultDone),
			ModifyRequest: (None, ModifyResponse),
			AddRequest: (None,  AddResponse),
			DelRequest: (None, DelResponse),
			ModifyDNRequest: (None, ModifyDNResponse),
			CompareRequest: (None, CompareResponse),
			AbandonRequest: (None, None),
			ExtendedRequest: (self.handle_extended, ExtendedResponse),
		}
		shallowmsg, rest = ShallowLDAPMessage.from_ber(data)
		if shallowmsg.protocolOp is None:
			print('Ignoring unknown message')
			return rest
		func, errfunc = handlers[shallowmsg.protocolOp]
		msg = None
		try:
			msg, _ = LDAPMessage.from_ber(data)
		except Exception as e:
			if errfunc:
				self.send_msg(LDAPMessage(shallowmsg.messageID, errfunc(LDAPResultCode.protocolError)))
			traceback.print_exc()
			return rest
		print('received', msg)
		try:
			if func:
				func(msg)
			elif errfunc:
				self.send_msg(LDAPMessage(shallowmsg.messageID, errfunc(LDAPResultCode.insufficientAccessRights)))
		except LDAPError as e:
			if errfunc:
				self.send_msg(LDAPMessage(shallowmsg.messageID, errfunc(e.code, diagnosticMessage=e.message)))
			return rest
		except Exception as e:
			if errfunc:
				self.send_msg(LDAPMessage(shallowmsg.messageID, errfunc(LDAPResultCode.other)))
			traceback.print_exc()
			return rest
		return rest

	def send_msg(self, msg):
		print('sending', msg)
		self.request.sendall(LDAPMessage.to_ber(msg))

	def handle(self):
		self.keep_running = True
		self.bind_object = None
		data = b''
		while self.keep_running:
			try:
				data = self.handle_message(data)
			except IncompleteBERError:
				chunk = self.request.recv(5)
				if not chunk:
					return
				data += chunk

